/* $Workfile: AudioVolume.c $       $Revision: 19 $      $Modtime: 3/14/08 3:22p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    AudioVolume.c

    Example of how a Volume Control function can be implemented using
    the following board combinations:
    
      Thor-1-PCI-Basic/Vidar-5x1-ASM-Basic
      Sleipnir-PCI v2.0

    DSP0 is loaded with a custom DSP application called AudioVolume.out (Vidar-5x1)
    or Dsp55AudioVolume.out (Sleipnir-PCI v2.0). 

    The volume range is 0 - 100, where 100 is max volume, and 0 is no 
    volume (the Codec is turned off).
    
    The AudioVolume.out application assumes the following timeslot connections
    in the incoming direction:

    TS0   A voice time slot the T1 or E1 span (in u-law or a-law)
    TS1   A voice time slot the T1 or E1 span (in u-law or a-law)
    TS2   A voice time slot the T1 or E1 span (in u-law or a-law)
    TS3   A voice time slot the T1 or E1 span (in u-law or a-law)

    The AudioVolume.out application assumes the following timeslot connections
    in the outgoing direction:

    TS0   Codec 0 Speaker  
    TS1   Codec 1 Speaker  
    TS2   Codec 2 Speaker  
    TS3   Codec 3 Speaker  

    For the Thor-1/Vidar-5x1 board combo the timeslot connections are performed
    by setting the sqTimeSlotMask in the OtxDrvConnectPipe() calls. 

    For the Sleipnir-PCI board the timeslots connections are performed by sending
    a "timeslot-connect" command (AV_IOCTL_SET_TS) to the DSP application.

    There are very few differences in this application for the two types of
    supported boards. To find the differences search for the following string:
    
      OTX_DEVICE_SLEIPNIR_1_PCI

    The applications uses a "key-command" interface. 

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"


#if 0
#include <AudioVolumeIo.h>   // IO Control codes for the DSP to host communication
#else 
#include <Dsp55AudioVolumeIo.h>   // IO Control codes for the DSP to host communication (OTX_DEVICE_SLEIPNIR_1_PCI)
#endif
#include <Dsp54Def.h>  // IO Control codes for the DSP to host communication
#include <DspIoctl.h>  // IO Control codes for the DSP to host communication

#define DEMO_TASK_REF 1L          // Task Reference for a non-blocking call
#define DEMO_TASK_REF_DEBUG 501   // Task Reference for a non-blocking call
#define DEMO_DSP_MAX 1
#define DEMO_NO_OF_LIS 1
#define DEMO_EVENT_DATA_BUF_SIZE 16

#define DEMO_DEFAULT_LOG_FILE "AudioVolume.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  OTX_BOOL        bExitOnError;
  char           *logFileName;
  char           *dspProgName;
  OTX_INT32       nNicBoardType;
  OTX_INT32       nAsmBoardType;
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);
static void DemoCheckDspRc(IN OTX_UINT16 nDspRc);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);

static void DemoConnectPipeEnds(
  IN OTX_HANDLE hPipe, 
  IN OTX_HANDLE hInpDev, 
  IN OTX_INT32  nInpTs, 
  IN OTX_HANDLE hOutDev, 
  IN OTX_INT32  nOutTs
);

static void DemoDisconnectPipeEnds(
  IN OTX_HANDLE hPipe
);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;


OsaMutexS g_demoLogfCritSec;


AudioVolumeIoInSetVolumeS  g_volume;
AudioVolumeIoInSetLawS     g_law;

OTX_INT32 g_sqCodecConnections[4]; // To keep track of which timeslot of the T1/E1 span that each code is connected to.

struct {
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hAsmController0;
  OTX_HANDLE hLi[1];
  OTX_HANDLE hDsp[1];
  OTX_HANDLE hCodec[4];
} g_hPhysDevs;


struct {
  OTX_HANDLE hAudioVolumeApp;     // Custom DSP App
} g_hLogicDevs;

struct {
  OTX_HANDLE hFromT1E1[4];           // Connects to 4 timeslots (selectable) from the T1 or E1 span to the DSP.
  OTX_BOOL   bT1E1Connected[4];      // True is connected
  OTX_HANDLE hFromAudioVolumeTs[4];  // The output from the AudioVolume. Connects to Codec speakers
  OTX_BOOL   bAVConnected[4];        // True is connected
} g_hPipes;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/

void DemoExit(void)
/*******************************************************************************
  @func
    Exit the program gracefully
*******************************************************************************/
{

  DemoLogf("Exiting...\n");

  OsaThreadSleep(1000); 

  OtxDrvReleaseLib();
  OsaConsoleExit();
  exit(1);
}



void DemoLoadVolumeValues(void)
/*******************************************************************************
  @func
    Load the gain values to the AudioVolume Application
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_UINT32 nBytesReturned;
  OTX_UINT16 nAudioVolumeRc;


  // Set volume for all outputs
  result = OtxDspIoControl(
    g_hLogicDevs.hAudioVolumeApp, 
    AV_IOCTL_SET_VOLUME, 
    &g_volume, sizeof(g_volume), 
    &nAudioVolumeRc, sizeof(nAudioVolumeRc), 
    &nBytesReturned, 
    OTX_TASK_SYNC, 
    OTX_TIME_INFINITY); 
  DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
  DemoCheckDspRc(nAudioVolumeRc);

  //DemoLogf("Sending law=%d to the DSP\n", g_law.law);
  // Set companding law
  result = OtxDspIoControl(
    g_hLogicDevs.hAudioVolumeApp, 
    AV_IOCTL_SET_LAW, 
    &g_law, sizeof(g_law), 
    &nAudioVolumeRc, sizeof(nAudioVolumeRc), 
    &nBytesReturned, 
    OTX_TASK_SYNC, 
    OTX_TIME_INFINITY); 
  DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
  DemoCheckDspRc(nAudioVolumeRc);

} 



void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Open the Physical Devices on the Thor-2-PCI board
  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller (Thor-1-PCI board or Thor-1-PCI-Basic)
    switch (g_demoEnv.nNicBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI/Thor-1-PCI NIC board\n");
    break;
  case OTX_DEVICE_THOR_1_PCI_BASIC:
    DemoLogf("Opening Thor-1-PCI-Basic NIC board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir NIC board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nNicBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nNicBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
  }


  // Get handles to Codec devices #0 - #3
  //DemoLogf("Setting Codecs to g_demoEnv.law=%d \n", g_demoEnv.law);
  for (i = 0; i < 4; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

    // Use G.711 U-law (North American) companded PCM data (8 bit samples) for T1 and A-law for E1
    result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.law); DemoCheckRc(g_hPhysDevs.hCodec[i], result);
  }

  if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // We will source the T1/E1 device (FALC) from the DSP device (not used on this board), and source the DSP from the T1E1 device
    //  T1E1(FALC) <-- DSP
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_FALC_SOURCED, 
                                           OTX_BRD_FALC_SOURCED_DSP); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    //  DSP <-- T1E1(FALC)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController0, 
                                           OTX_ATTR_BRD_DSP_SOURCED, 
                                           OTX_BRD_DSP_SOURCED_FALC); 
    DemoCheckRc(g_hPhysDevs.hBoardController0, result);

#if 1
    // Optional: It is not neccessary to open Now when the T1E1 device is initialized we can open the QuadCodec device. This is a good way to verify that the Codec device is properly activated by the driver
    if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
      OTX_HANDLE hQuadCodec;
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_QUAD_CODEC, 0, 0, g_hEventQueue, &hQuadCodec); DemoCheckRc(g_hPhysDevs.hBoardController0, result); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }
#endif

    // The Sleipnir does not have an ASM board to host the DSP. The NIC board itself is the host for the DSP
    g_hPhysDevs.hAsmController0 = OTX_INVALID_HANDLE_VALUE;
    for (i = 0; i < DEMO_DSP_MAX; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    }


  } else {
    // Get a handle to to the ASM Daughter Board
    DemoLogf("Opening %s ASM board\n", g_demoEnv.nAsmBoardType == OTX_DEVICE_VIDAR_5x_ASM ? "Vidar-5x4-ASM" : "Vidar-5x1-ASM-Basic");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, (OtxPhysicalDeviceTypeE) g_demoEnv.nAsmBoardType, 0, 0, g_hEventQueue, &g_hPhysDevs.hAsmController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    
    for (i = 0; i < DEMO_DSP_MAX; i++) {
      DemoLogf("Opening DSP#%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
    }
  }
  

  // Load DSP Programs
  i = 0;
  DemoLogf("Loading %s program to DSP#%d\n", g_demoEnv.dspProgName, i);
  result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.dspProgName, "_c_int00"); 
  if (OTX_RESULT_CODE(result) == OTX_E_FAIL) {
    DemoLogf("*** Error loading the DSP program. Is the '%s' file accessible?\n", g_demoEnv.dspProgName);
    DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  } else if (OTX_RESULT_CODE(result) != OTX_S_OK) {
    DemoLogf("*** Error loading the %s program. Possibly an SDK Lib version mismatch.\n", g_demoEnv.dspProgName);
    DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eRxLineCode                = OTX_T1_LC_AMI;   // AMI, B8ZS
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;    // F12 (D3/4 or SF), F4, ESF, F72 (SLC96)
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;  
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_FALSE;
  pT1CfgData->bfClearChannels            = 0x000000;
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  i = 0;

  // Set termination for the T1/E1 Front End
  result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
  result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
    // T1 mode
    result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  } else {
    // E1 mode
    result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }
  result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // OTX_DEVICE_SLEIPNIR_1_PCI The Sleipnir does not have an ASM board to host the DSP. The NIC board itself is the host for the DSP.
  if (g_hPhysDevs.hAsmController0 != OTX_INVALID_HANDLE_VALUE) {
    result = OtxDrvEnable(g_hPhysDevs.hAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hAsmController0, result);
  }

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  result = OtxDrvEnable(g_hPhysDevs.hLi[0], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);

}



void DemoCreateLogicalDevices(void)
/*******************************************************************************
  @func
    Create the needed Logical Devices
  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  // DSP0
  i = 0; 
  // Create one AudioVolume Application Logical Device in DSP#0
  result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, 0, &g_hLogicDevs.hAudioVolumeApp); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  // Empty. The logial devices do not need initalization in the application.
}




void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;

  result = OtxDrvEnable(g_hLogicDevs.hAudioVolumeApp, OTX_TASK_SYNC, 5000L); 
  DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

}


/*******************************************************************************
  @func
    Connects a Codec to a specific timeslot on the T1E1 span. 
*******************************************************************************/
void DemoConnectCodec(
  OTX_INT32 nCodec,  // 0-3
  OTX_INT32 nTs      // 0-31  (really 1-31 for E1 and 1-24 for T1)
  
)
{
#if 0
  /*
  if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // Since there is not Time-space switch on the Sleipnir board we will send   
    // a command to the DSP to instruct to pick up a certain timeslot
    OTX_RESULT result;
    OTX_UINT32 nBytesReturned;
    OTX_UINT16 nAudioVolumeRc;
    AudioVolumeIoInSetTsS sIn;
    
    DemoLogf("Codec%d  Ts=%d\n", nCodec, nTs);
    
    sIn.nCodec = (short)nCodec;
    sIn.nTs    = (short)nTs;
    
    // Set the timeslot for one Codec channel
    result = OtxDspIoControl(
      g_hLogicDevs.hAudioVolumeApp, 
      AV_IOCTL_SET_TS, 
      &sIn, sizeof(sIn), 
      &nAudioVolumeRc, sizeof(nAudioVolumeRc), 
      &nBytesReturned, 
      OTX_TASK_SYNC, 
      OTX_TIME_INFINITY); 
    DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
    DemoCheckDspRc(nAudioVolumeRc);


  } else {
    DemoConnectPipeEnds(
      g_hPipes.hFromT1E1[nCodec], 
      g_hPhysDevs.hLi[0],  nTs,    
      g_hPhysDevs.hDsp[0], nCodec   // to AudioVolume application. Fixed timeslot
      );
  }

  g_sqCodecConnections[nCodec] = nTs;

  */
#else
    if(g_hPipes.bT1E1Connected[nCodec]) {
      DemoLogf("   Pipe already connected\n");
    }
    else {
      g_hPipes.bT1E1Connected[nCodec] = OTX_TRUE;
      DemoConnectPipeEnds(
        g_hPipes.hFromT1E1[nCodec], 
        g_hPhysDevs.hLi[0],  nTs,    
        g_hPhysDevs.hDsp[0], nCodec   // to AudioVolume application. Fixed timeslot
        );

      g_sqCodecConnections[nCodec] = nTs;
    }

#endif

}

/*******************************************************************************
  @func
    Disconnects a Codec from the T1E1 span. 
*******************************************************************************/
void DemoDisconnectCodec(
  OTX_INT32 nCodec  // 0-3
)
{
  /*
  if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // Nothing
  } else {
  */
  if(g_hPipes.bT1E1Connected[nCodec] == OTX_FALSE) {
    DemoLogf("   Pipe not connected\n");
  }
  else {
    g_hPipes.bT1E1Connected[nCodec] = OTX_FALSE;
    DemoDisconnectPipeEnds(g_hPipes.hFromT1E1[nCodec]);
  }
}



/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.
  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
void DemoConnectDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;
  
  // Create voice pipes 
  for (i = 0; i < 4; i++) {
    result = OtxDrvCreatePipe(64, &g_hPipes.hFromT1E1[i]); 
    DemoCheckRc(g_hPipes.hFromT1E1[i], result);
  }
  
  for (i = 0; i < 4; i++) {
    result = OtxDrvCreatePipe(64, &g_hPipes.hFromAudioVolumeTs[i]); 
    DemoCheckRc(g_hPipes.hFromAudioVolumeTs[i], result);
  }

  // Note: Timeslot mapping in the T1E1 device 
  // E1 mode: TS0 - TS31 from the E1 span (index 0 is TS0, the synchronization channel)
  // T1 mode: TS1 - TS24 from the T1 span (index 0, and index 18-24 are unused)

  // Input pipes to DSP0 (AudioVolume)
  // These connections can be modified interactively by keyboard input commands.
  DemoConnectCodec(0, g_sqCodecConnections[0]);
  DemoConnectCodec(1, g_sqCodecConnections[1]);
  DemoConnectCodec(2, g_sqCodecConnections[2]);
  DemoConnectCodec(3, g_sqCodecConnections[3]);
  
  // Output pipes from DSP0 (AudioVolume). 
  // These connections are fixed and do not change in this application.
  // TS0   Codec 0 Speaker 
  DemoLogf("DemoConnectDevices Audio Volume TS0 Codec 0 TS1\n");
  g_hPipes.bAVConnected[0] = OTX_TRUE;
  DemoConnectPipeEnds(
    g_hPipes.hFromAudioVolumeTs[0], 
    g_hPhysDevs.hDsp[0], 0,   // DSP0 AUDIOVOLUME TS0
    g_hPhysDevs.hCodec[0], 1  // to Codec0 Speaker (Note that Codec0 uses TS1 on the codec highway)
    );
#if 0
  DemoLogf("DemoConnectDevices Connect logical device Codec 0\n");
  result = OtxDrvConnectLogicalDevice(  // Probably a redundant call. Connections are hardcoded in AUDIOVOLUME.
    g_hLogicDevs.hAudioVolumeApp, -1,
    g_hPipes.hFromAudioVolumeTs[0], -1
    ); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif
  
  // TS1   Codec 1 Speaker 
  DemoLogf("DemoConnectDevices Audio Volume TS1 Codec 1 TS2\n");
  g_hPipes.bAVConnected[1] = OTX_TRUE;
  DemoConnectPipeEnds(
    g_hPipes.hFromAudioVolumeTs[1], 
    g_hPhysDevs.hDsp[0], 1,   // DSP0 AUDIOVOLUME TS1
    g_hPhysDevs.hCodec[1], 2  // to Codec1 Speaker (Note that Codec1 uses TS2 on the codec highway)
    );
#if 0
  DemoLogf("DemoConnectDevices Connect logical device Codec 1\n");
  result = OtxDrvConnectLogicalDevice(  // Probably a redundant call. Connections are hardcoded in AUDIOVOLUME.
    g_hLogicDevs.hAudioVolumeApp, -1,
    g_hPipes.hFromAudioVolumeTs[1], -1
    ); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif
  
  // TS2   Codec 2 Speaker 
  DemoLogf("DemoConnectDevices Audio Volume TS2 Codec 2 TS3\n");
  g_hPipes.bAVConnected[2] = OTX_TRUE;
  DemoConnectPipeEnds(
    g_hPipes.hFromAudioVolumeTs[2], 
    g_hPhysDevs.hDsp[0], 2,   // DSP0 AUDIOVOLUME TS2
    g_hPhysDevs.hCodec[2], 3  // to Codec2 Speaker (Note that Codec2 uses TS3 on the codec highway)
    );
#if 0
  DemoLogf("DemoConnectDevices Connect logical device Codec 2\n");
  result = OtxDrvConnectLogicalDevice(  // Probably a redundant call. Connections are hardcoded in AUDIOVOLUME.
    g_hLogicDevs.hAudioVolumeApp, -1,
    g_hPipes.hFromAudioVolumeTs[2], -1
    ); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif
  
  // TS3   Codec 3 Speaker 
  DemoLogf("DemoConnectDevices Audio Volume TS3 Codec 3 TS4\n");
  g_hPipes.bAVConnected[3] = OTX_TRUE;
  DemoConnectPipeEnds(
    g_hPipes.hFromAudioVolumeTs[3], 
    g_hPhysDevs.hDsp[0], 3,   // DSP0 AUDIOVOLUME TS3
    g_hPhysDevs.hCodec[3], 4  // to Codec3 Speaker (Note that Codec3 uses TS4 on the codec highway)
    );
#if 0
  DemoLogf("DemoConnectDevices Connect logical device Codec 3\n");
  result = OtxDrvConnectLogicalDevice(  // Probably a redundant call. Connections are hardcoded in AUDIOVOLUME.
    g_hLogicDevs.hAudioVolumeApp, -1,
    g_hPipes.hFromAudioVolumeTs[3], -1
    ); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif
  
  
}


/*******************************************************************************
  @func
    Connects both ends of a pipe. 
*******************************************************************************/
static void DemoConnectPipeEnds(
  IN OTX_HANDLE hPipe, 
  IN OTX_HANDLE hInpDev, 
  IN OTX_INT32  nInpTs, 
  IN OTX_HANDLE hOutDev, 
  IN OTX_INT32  nOutTs
)
{
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  
  if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // There is no time-space switch on this board. The timeslot connections to the Codec are handled by the DSP program
    //return;
  } 
    
  DemoLogf("DemoConnectPipeEnds hPipe(0x%X)\n", hPipe);
  DemoLogf("                    hInpDev(0x%X) nInpTs(%d)\n", hInpDev, nInpTs);
  DemoLogf("                    hOutDev(0x%X) nOutTs(%d)\n", hOutDev, nOutTs);
  DemoLogf("\n");

  // Connect input end of the pipe
  memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
  sqTimeSlotMask[nInpTs] = 0xFF;
  
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_INPUT, hInpDev, -1, sqTimeSlotMask, 32);
  DemoCheckRc(hPipe, result);
  
  // Connect output end of the pipe
  memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
  sqTimeSlotMask[nOutTs] = 0xFF;
  
  result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, hOutDev, -1, sqTimeSlotMask, 32);
  DemoCheckRc(hPipe, result);

}


/*******************************************************************************
  @func
    Disconnects both ends of a pipe. 
*******************************************************************************/
static void DemoDisconnectPipeEnds(
  IN OTX_HANDLE hPipe 
)
{
  OTX_RESULT result;

  //if (g_demoEnv.nNicBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // There is no time-space switch on this board. 
    // The timeslot connections to the Codec are handled by the DSP program
    //return;
  //} 

  DemoLogf("DemoDisconnectPipeEnds hPipe(0x%X) INPUT\n", hPipe);

  // Disconnect input end of the pipe
  result = OtxDrvDisconnectPipe(hPipe, OTX_PIPE_INPUT);
  DemoCheckRc(hPipe, result);

  DemoLogf("DemoDisconnectPipeEnds hPipe(0x%X) OUTPUT\n", hPipe);

  // Disconnect output end of the pipe
  result = OtxDrvDisconnectPipe(hPipe, OTX_PIPE_OUTPUT);
  DemoCheckRc(hPipe, result);

}






static void DemoCheckDspRc(IN OTX_UINT16 nDspRc)
/*******************************************************************************
  @func
    Check the result from a call to a DSP Ioctl call. If the result is 
    a failure code then prompt the user if he wants to exit.
*******************************************************************************/
{
  short c;

  if (nDspRc != OTXDSP_SDK_S_OK) {

    DemoLogf("DSP Application Error Code: %.X) \n", nDspRc);
    
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
    if (c == 'y' || c == 'Y') {
      // Clean up
      DemoExit();
    }

  }
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");

    do {
      // Multiple reads of events
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }

    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];


  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
    break;

  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the AudioVolume.out logical device
    if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
      OTX_UINT8  buf[8196];
      OTX_UINT32 nBytesReturned;
      OTX_UINT32 nIoControlCode;

      otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8196, &nBytesReturned, &nIoControlCode);
      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        switch(nIoControlCode) {
        case AV_NOTIFY_DEBUG_DATA:
          {
            AudioVolumeNotifyGetDebugDataS *pDataS = (AudioVolumeNotifyGetDebugDataS *)buf;

            DemoPrintTimeStamp(pEventData->m_nTimeStamp);
            if (nBytesReturned != sizeof(AudioVolumeNotifyGetDebugDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              DemoLogf("AV_NOTIFY_DEBUG_DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              DemoLogf("nData0 = 0x%.4X  (%+6d)\n", pDataS->nData0, (short)pDataS->nData0);
              DemoLogf("nData1 = 0x%.4X  (%+6d)\n", pDataS->nData1, (short)pDataS->nData1);
              DemoLogf("nData2 = 0x%.4X  (%+6d)\n", pDataS->nData2, (short)pDataS->nData2);
              DemoLogf("nData3 = 0x%.4X  (%+6d)\n", pDataS->nData3, (short)pDataS->nData3);
              DemoLogf("nData4 = 0x%.4X  (%+6d)\n", pDataS->nData4, (short)pDataS->nData4);
              DemoLogf("nData5 = 0x%.4X  (%+6d)\n", pDataS->nData5, (short)pDataS->nData5);
              DemoLogf("nData6 = 0x%.4X  (%+6d)\n", pDataS->nData6, (short)pDataS->nData6);
              DemoLogf("nData7 = 0x%.4X  (%+6d)\n", pDataS->nData7, (short)pDataS->nData7);
              //DemoLogf("nData34= %+d\n", ((long)pDataS->nData4 << 16) + pDataS->nData3);
            }
          }
          break;
        case AV_NOTIFY_STREAM_RX_DATA:
          {
            if (nBytesReturned != sizeof(AudioVolumeNotifyStreamRxDataS)) {
              DemoLogf("Unexpected data length for this IoControlCode (nIoControlCode = %d, nBytesReturned=%d)\n", nIoControlCode, nBytesReturned);
            } else {
              static OTX_INT32 i = 0;

              if ((i++ % 500) == 0) {
                DemoLogf("RX DATA[%d] (nRequestId=%d, nBytes=%d)\n", pEventData->m_nSourceId, pEventData->m_nRequestId, nBytesReturned);
              }
#if 1         
              {
                AudioVolumeNotifyStreamRxDataS *pDataS = (AudioVolumeNotifyStreamRxDataS *)buf;
                for (i = 0; i < 25; i++) {
                  DemoLogf("%+6d\n", (short)pDataS->nData[i]);
                }
                //DemoLogf("\n");
              }
#endif        
            }
          }
          break;
        default:
          DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
          break;
        }
      } else {
        OTX_RESULT errorCode = otxRc;
        DemoPrintTimeStamp(pEventData->m_nTimeStamp);
        otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
        } else {
          DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
        }
      }

    } else {

      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
      } else {
        DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
      }
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}



OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  static OTX_BOOL bStreamRxData;
  static OTX_INT32 nCurrentCodec = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-3 Select current Codec. Current Codec is [%d]\n", nCurrentCodec);
    DemoLogf(" c   Connect current Codec to specific T1E1 timeslot. Current TS is %d\n", g_sqCodecConnections[nCurrentCodec]);
    DemoLogf(" l   Query Line Status of the Li\n");
    DemoLogf(" L   Query Line Status of the Li (more detailed information)\n");
    DemoLogf(" >   Increase volume by 1 for current Codec\n");
    DemoLogf(" <   Decrease volume by 1 for current Codec\n");
    DemoLogf(" A   Set 6.02 dB loss (attenuation) for all Codecs\n");
    DemoLogf(" a   Set 0 dB loss (attenuation) for all Codecs\n");
    DemoLogf(" m   Connect pipe to codec T1E1 to DSP \n");
    DemoLogf(" M   Disconnect pipe from codec T1E1 to DSP \n");
    DemoLogf(" p   Connect pipe to codec Audio Volume\n");
    DemoLogf(" P   Disconnect pipe from codec Audio Volume\n");

    DemoLogf("\nThe following commands are mainly for debugging:\n");
    DemoLogf(" K   Get Expanded sample value (16-bits) from the DSP User Application\n");
    DemoLogf(" e   Enable AudioVolume application\n");
    DemoLogf(" E   Disable AudioVolume application\n");
    DemoLogf(" R   Send RESET command to AudioVolume application\n");
    DemoLogf(" D   Get Debug data from the DSP User Application\n");
    DemoLogf(" T   Enable/Disable (toggle) continous stream of the received data stream\n");
    DemoLogf(" f   Get full frame of the received data stream\n");

    DemoLogf("\nCurrent Codec is %d. Volume is %d. Connected to TS%d\n", nCurrentCodec, g_volume.sqCodec[nCurrentCodec], g_sqCodecConnections[nCurrentCodec]);
    break;

  case 'c':
    {
      char strBuf[12];
      OTX_INT32 nTs = 0;
      DemoLogf("Codec%d is currently connected to TS%d. Please enter new TS: ", nCurrentCodec, g_sqCodecConnections[nCurrentCodec]);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTs = strtol(strBuf, NULL, 0);
      }

      if ((nTs < 0) || (nTs > 31)) {
        DemoLogf("TS out of range (%d). Valid range is 0-31. Command aborted.\n", nTs);
        break;
      }

      DemoDisconnectCodec(nCurrentCodec);
      DemoConnectCodec(nCurrentCodec, nTs);
        
      DemoLogf("Done. Codec%d is now connected to TS%d.\n", nCurrentCodec, nTs);
    }
    break;

  case '<':
    {
      OTX_RESULT result;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nAudioVolumeRc;

      if (g_volume.sqCodec[nCurrentCodec] > 0) {
        // Decrease volume by one
        g_volume.sqCodec[nCurrentCodec]--;
      }
      
      DemoLogf("Codec%d  Volume=%d\n", nCurrentCodec, g_volume.sqCodec[nCurrentCodec]);
      
      // Set volume for all outputs
      result = OtxDspIoControl(
        g_hLogicDevs.hAudioVolumeApp, 
        AV_IOCTL_SET_VOLUME, 
        &g_volume, sizeof(g_volume), 
        &nAudioVolumeRc, sizeof(nAudioVolumeRc), 
        &nBytesReturned, 
        OTX_TASK_SYNC, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      DemoCheckDspRc(nAudioVolumeRc);
    }
    break;
    
    
  case '>':
    {
      OTX_RESULT result;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nAudioVolumeRc;
      
      if (g_volume.sqCodec[nCurrentCodec] < 100) {
        // Increase volume by one
        g_volume.sqCodec[nCurrentCodec]++;
      }
      
      DemoLogf("Codec%d  Volume=%d\n", nCurrentCodec, g_volume.sqCodec[nCurrentCodec]);
      
      // Set volume for all outputs
      result = OtxDspIoControl(
        g_hLogicDevs.hAudioVolumeApp, 
        AV_IOCTL_SET_VOLUME, 
        &g_volume, sizeof(g_volume), 
        &nAudioVolumeRc, sizeof(nAudioVolumeRc), 
        &nBytesReturned, 
        OTX_TASK_SYNC, 
        OTX_TIME_INFINITY); 
      DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      DemoCheckDspRc(nAudioVolumeRc);

    }
    break;
  
  case '0':
    nCurrentCodec = 0;
    DemoLogf("Codec%d is now current codec\n", nCurrentCodec);
    break;
        
  case '1':
    nCurrentCodec = 1;
    DemoLogf("Codec%d is now current codec\n", nCurrentCodec);
    break;
        
  case '2':
    nCurrentCodec = 2;
    DemoLogf("Codec%d is now current codec\n", nCurrentCodec);
    break;
        
  case '3':
    nCurrentCodec = 3;
    DemoLogf("Codec%d is now current codec\n", nCurrentCodec);
    break;
        
  case 'A':
    {
      OTX_INT32 nCmd;
      OTX_INT32 nCh;
      OTX_INT32 nReg = 0x50; // Analog Gain Register
      OTX_INT32 nData = 0x20;  // 0x20 = 6.02 dB loss

      DemoLogf("Setting 6.02 dB loss (attenuation) for all Codecs\n");

      for (nCh = 0; nCh < 4; nCh++) {
        nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
        OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);
      }

      DemoLogf("Done.\n");

    }
    break;

  case 'a':
    {
      OTX_INT32 nCmd;
      OTX_INT32 nCh;
      OTX_INT32 nReg = 0x50; // Analog Gain Register
      OTX_INT32 nData = 0x00;  // 0x20 = 0 dB loss

    DemoLogf("Setting 0 dB loss (attenuation) for all Codecs\n");

      for (nCh = 0; nCh < 4; nCh++) {
        nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
        OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);
      }

      DemoLogf("Done.\n");

    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 1; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'L':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 1; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi[nLi], &sLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;


  case 'B': // Just for debugging
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < DEMO_DSP_MAX; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;


  case 'T': // Just for debugging
    {
      AudioVolumeIoInStreamRxDataS sIn;
      OTX_UINT32 nBytesReturned;
      OTX_UINT16 nAudioVolumeRc;
      
      bStreamRxData = !bStreamRxData;
      
      DemoLogf("%s stream of RX data for User App ...", bStreamRxData ? "Enabling" : "Disabling");
      
      sIn.bStreamRxData = (OTX_INT16) bStreamRxData;
      
      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_STREAM_RX_DATA, &sIn, sizeof(sIn), &nAudioVolumeRc, sizeof(nAudioVolumeRc), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      
      DemoLogf("Done (nAudioVolumeRc=%d, nBytesReturned=%d).\n", nAudioVolumeRc, nBytesReturned);
    }
    break;
  
  case 'e':  // Just for debugging
    {
      DemoLogf("Enable AudioVolume App ...");
      result = OtxDrvEnable(g_hLogicDevs.hAudioVolumeApp,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      DemoLogf("Done.\n");
    }
    break;

  case 'E':  // Just for debugging
    {
      DemoLogf("Disable AudioVolume App ...");
      result = OtxDrvDisable(g_hLogicDevs.hAudioVolumeApp,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      DemoLogf("Done.\n");
    }
    break;

  case 'R':
    {
      DemoLogf("Send RESET Command to AudioVolume App ...");
      result = OtxDrvReset(g_hLogicDevs.hAudioVolumeApp); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);
      DemoLogf("Done.\n");
    }
    break;

  case 'D':  // Just for debugging
    {
      OTX_UINT32 nBytesReturned;
      DemoLogf("Get Debug Data for AudioVolume App (non-blocking call)...");

      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_GET_DEBUG_DATA, NULL, 0, NULL, 0, &nBytesReturned, DEMO_TASK_REF_DEBUG, 5000); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case 'K':  // Just for debugging
    {
      OTX_UINT32 nBytesReturned;
      AudioVolumeIoOutGetSampleS dataS;

      DemoLogf("Get sample value (16-bit) from AudioVolume App \n");

      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_GET_SAMPLE, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(AudioVolumeIoOutGetSampleS)) {
          DemoLogf("Unexpected data length for AV_IOCTL_GET_SAMPLE (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("AudioVolume Error Code: %d\n", dataS.nResult);
        } else {
          DemoLogf("DataSample = %+.6d\n", (short)dataS.nDataSample);
        }
      }

      DemoLogf("Done (nBytesReturned=%d).\n", nBytesReturned);
    }
    break;

  case '{':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      static OTX_INT32 nCh = 0;
      static OTX_INT32 nReg = 0x70;
      static OTX_INT32 nData;
      
      DemoLogf("Read Register from Codec.\n");
      DemoLogf("Enter Channel[%d]: ", nCh);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nCh = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      DemoLogf("Enter Register[0x%.2X]: ", nReg);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nReg = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      
      nCmd = 0x3010 | nCh | (nReg << 24);
      DemoLogf("Reading Codec Ch=%d, Register=0x%.2X\n", nCh, nReg);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);
    }
    break;

  case '}':
    {
      char szBuf[64];
      OTX_INT32 nCmd;
      static OTX_INT32 nCh = 0;
      static OTX_INT32 nReg = 0x70;
      static OTX_INT32 nData = 0x00;
      static OTX_INT32 nData2 = 0x00;
      
      DemoLogf("Write Register to Codec.\n");
      DemoLogf("Enter Channel[%d]: ", nCh);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nCh = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      DemoLogf("Enter Register[0x%.2X]: ", nReg);
      OsaConsoleGetStr(szBuf);
      if (szBuf[0]) {
        nReg = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      if ((nReg >=0x80) && (nReg <=0x82)) {
        // Dual data byte command
        DemoLogf("Enter Data1[0x%.2X]: ", nData);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData = (OTX_INT32) strtol(szBuf, NULL, 0);
        }
        DemoLogf("Enter Data2[0x%.2X]: ", nData2);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData2 = (OTX_INT32) strtol(szBuf, NULL, 0);
        }
        nCmd = 0x3020 | nCh | (nData << 24) | (nData2 << 16);
        if (nReg == 0x82) nCmd |= 0x10;
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data1=0x%.2X, Data2=0x%.2X\n", nCh, nReg, nData, nData2);
      } else {
        // Single data byte command
        DemoLogf("Enter Data[0x%.2X]: ", nData);
        OsaConsoleGetStr(szBuf);
        if (szBuf[0]) {
          nData = (OTX_INT32) strtol(szBuf, NULL, 0);
        }

        nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
        DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      }
      
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);
    }
    break;


  case '|':
    {
      OTX_INT32 nCmd;
      OTX_INT32 nTs = 5;
      static OTX_INT32 nCh = 0;
      static OTX_INT32 nReg = 0x70;
      static OTX_INT32 nData = 0x00;
      static OTX_INT32 nData2 = 0x00;

      DemoDisconnectCodec(nCh);
      DemoConnectCodec(nCh, nTs);
        
      DemoLogf("Codec%d is now connected to TS%d.\n", nCh, nTs);

      // Deactivate ch1
      nCh = 1;
      nReg = 0x00; 
      nData = 0x00;
      nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
      DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);

      // Deactivate ch2
      nCh = 2;
      nReg = 0x00; 
      nData = 0x00;
      nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
      DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);

      // Deactivate ch3
      nCh = 3;
      nReg = 0x00; 
      nData = 0x00;
      nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
      DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);

      // Deactivate ch3
      nCh = 3;
      nReg = 0x00; 
      nData = 0x00;
      nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
      DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);

      // TSA loopback ch0
      nCh = 0;
      nReg = 0x70; 
      nData = 0x04;
      nCmd = 0x3000 | nCh | (nReg << 24) | (nData << 16);
      DemoLogf("Writing Codec Ch=%d, Register=0x%.2X, Data=0x%.2X\n", nCh, nReg, nData);
      OtxDrvDbgDump(g_hPhysDevs.hBoardController0, nCmd);

    }
    break;


  case 'f':
    {
      OTX_UINT32 nBytesReturned;
      AudioVolumeIoOutGetFullFrameS dataS;

      DemoLogf("\nT1/E1 RX highway timeslot data:\n");

      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_GET_FULL_RX_FRAME, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(AudioVolumeIoOutGetFullFrameS)) {
          DemoLogf("Unexpected data length for AV_IOCTL_GET_FULL_RX_FRAME (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          OTX_INT32 i;
          for (i = 0; i < 32; i++) {
            if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
            DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
          }
          DemoLogf("\n");

        }
      }


      DemoLogf("\nCodec TX highway timeslot data (TS1-TS4 represents the 4 Codec channels):\n");

      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_GET_FULL_TX_FRAME2, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(AudioVolumeIoOutGetFullFrameS)) {
          DemoLogf("Unexpected data length for AV_IOCTL_GET_FULL_TX_FRAME2 (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          OTX_INT32 i;
          for (i = 0; i < 32; i++) {
            if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
            DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
          }
          DemoLogf("\n");

        }
      }

      DemoLogf("\nCodec RX highway timeslot data (TS1-TS4 represents the 4 Codec channels):\n");

      result = OtxDspIoControl(g_hLogicDevs.hAudioVolumeApp, AV_IOCTL_GET_FULL_RX_FRAME2, NULL, 0, &dataS, sizeof(dataS), &nBytesReturned, OTX_TASK_SYNC, OTX_TIME_INFINITY); DemoCheckRc(g_hLogicDevs.hAudioVolumeApp, result);

      if (OTX_SUCCESS(result)) {
        if (nBytesReturned != sizeof(AudioVolumeIoOutGetFullFrameS)) {
          DemoLogf("Unexpected data length for AV_IOCTL_GET_FULL_RX_FRAME2 (nBytesReturned=%d)\n", nBytesReturned);
        } else if (dataS.nResult != OTXDSP_SDK_S_OK) {
          DemoLogf("Demo2 Error Code: %d\n", dataS.nResult);
        } else {
          OTX_INT32 i;
          for (i = 0; i < 32; i++) {
            if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
            DemoLogf("[%.3d]:%.2X ", i, (OTX_UINT8)dataS.nFrame[i]);
          }
          DemoLogf("\n");

        }
      }
    }
    break;

  case 'm':

    DemoLogf("Connect pipe T1E1 to DSP codec %d\n", nCurrentCodec);

    if(g_hPipes.bT1E1Connected[nCurrentCodec]) {
      DemoLogf("   Pipe already connected\n");
    }
    else {
      DemoConnectCodec(nCurrentCodec, g_sqCodecConnections[nCurrentCodec]);
/*
      DemoConnectPipeEnds(
        g_hPipes.hFromT1E1[nCurrentCodec], 
        g_hPhysDevs.hLi[0],
        g_sqCodecConnections[nCurrentCodec],   // DSP0 AUDIOVOLUME TS1
        g_hPhysDevs.hDsp[0], 
        nCurrentCodec  // to Codec1 Speaker (Note that Codec1 uses TS2 on the codec highway)
        );
*/
    }
    break;

  case 'M':

    DemoLogf("Disconnect pipe T1E1 to DSP codec %d\n", nCurrentCodec);

    if(g_hPipes.bT1E1Connected[nCurrentCodec] == OTX_FALSE) {
      DemoLogf("   Pipe not connected\n");
    }
    else {
      DemoDisconnectCodec(nCurrentCodec);

      //DemoDisconnectPipeEnds(g_hPipes.hFromT1E1[nCurrentCodec]);
    }
    break;

  case 'p':

    DemoLogf("Connect pipe to codec %d Audio Volume\n", nCurrentCodec);

    if(g_hPipes.bAVConnected[nCurrentCodec]) {
      DemoLogf("   Pipe already connected\n");
    }
    else {

      g_hPipes.bAVConnected[nCurrentCodec] = OTX_TRUE;

      DemoConnectPipeEnds(
        g_hPipes.hFromAudioVolumeTs[nCurrentCodec], 
        g_hPhysDevs.hDsp[0],
        nCurrentCodec,   // DSP0 AUDIOVOLUME TS1
        g_hPhysDevs.hCodec[nCurrentCodec], 
        nCurrentCodec + 1  // to Codec1 Speaker (Note that Codec1 uses TS2 on the codec highway)
        );
    }
    break;

  case 'P':

    DemoLogf("Disconnect pipe form codec %d Audio Volume\n", nCurrentCodec);

    if(g_hPipes.bAVConnected[nCurrentCodec] == OTX_FALSE) {
      DemoLogf("   Pipe not connected\n");
    }
    else {

      g_hPipes.bAVConnected[nCurrentCodec] = OTX_FALSE;

      DemoDisconnectPipeEnds(g_hPipes.hFromAudioVolumeTs[nCurrentCodec]);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *pEnv)
/*******************************************************************************
  @func
    Parse the program's command line arguments and calls upon
    requested actions.

    There are some optional switches that can be used following
    the driver name on the command line. Switches are are
    separated by spaces, and can appear in arbitrary order.
    Switches consist of a hyphen and a letter:
*******************************************************************************/
{
  short  i;

  // Set reasonable defaults before reading from command line
  pEnv->bExitOnError  = OTX_FALSE;
  pEnv->logFileName   = DEMO_DEFAULT_LOG_FILE;
  pEnv->nNicBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
  pEnv->nAsmBoardType = OTX_DEVICE_VIDAR_5X1_ASM_BASIC;
  pEnv->boardNo       = DEMO_DEFAULT_BOARD_NO;
  pEnv->law           = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  pEnv->eT1E1Mode     = OTX_T1E1_MODE_E1;
  pEnv->dspProgName   = "Dsp55AudioVolume.out";

  // struct AudioVolumeIoInSetVolumeS  // Set volume values for all AudioVolume output channels
  for (i = 0; i < 4; i++) {
    g_volume.sqCodec[i] = 50;  // Volume can be set between 0 and 100. 0 turns the sound completely off   
  }

  g_sqCodecConnections[0] = 1;  // This is arbitrary. Codec 0 connects to TS1 at startup
  g_sqCodecConnections[1] = 2;  // This is arbitrary. Codec 1 connects to TS2 at startup
  g_sqCodecConnections[2] = 3;  // This is arbitrary. Codec 2 connects to TS3 at startup
  g_sqCodecConnections[3] = 4;  // This is arbitrary. Codec 3 connects to TS4 at startup

  g_law.law = OTXDSP_ALAW;


  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        pEnv->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'N') {   // Board number
        pEnv->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'T') {   // Nic Board Type
        if ( *(argv[i]+2) == '1') {
          pEnv->nNicBoardType = OTX_DEVICE_THOR_PCI;
          pEnv->dspProgName   = "AudioVolume.out";
        } else if ( *(argv[i]+2) == 'B') {
          pEnv->nNicBoardType  = OTX_DEVICE_THOR_1_PCI_BASIC;
          pEnv->dspProgName   = "AudioVolume.out";
        } else if ( *(argv[i]+2) == 'S') {
          pEnv->nNicBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          pEnv->dspProgName   = "Dsp55AudioVolume.out";
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'A') {   // Asm Board Type
        if ( *(argv[i]+2) == '4') {
          pEnv->nAsmBoardType = OTX_DEVICE_VIDAR_5x_ASM;
        } else if ( *(argv[i]+2) == 'B') {
          pEnv->nAsmBoardType  = OTX_DEVICE_VIDAR_5X1_ASM_BASIC;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'M') {   // Li Mode
        if (*(argv[i]+2) == 'E') {   // E1 
          pEnv->eT1E1Mode    = OTX_T1E1_MODE_E1;
          pEnv->law          = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          g_law.law          = OTXDSP_ALAW;
        } else if (*(argv[i]+2) == 'T') {   // T1
          pEnv->eT1E1Mode    = OTX_T1E1_MODE_T1;
          pEnv->law          = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          g_law.law          = OTXDSP_ULAW;
        } else if (*(argv[i]+2) == 'X') {   // Debug mode
          pEnv->eT1E1Mode    = OTX_T1E1_MODE_E1;
          pEnv->law          = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          g_law.law          = 2;
        } else {
          DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
					exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
      exit(1);
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /N<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[1|B]        : Board Type used for the demo, where\n");
  DemoLogf("                    1 = Use Thor-2-PCI/Thor-1-PCI Board\n");
  DemoLogf("                    B = Use Thor-1-PCI-Basic Board\n");
  DemoLogf("                    C = Use Sleipnir-1-PCI 2.0 Board (default)\n");
  DemoLogf(" /A[4|B]        : Board Type used for the demo, where\n");
  DemoLogf("                    4 = Use Vidar-5x4-ASM Board\n");
  DemoLogf("                    B = Use Vidar-5x1-ASM-Basic Board\n");
  DemoLogf(" /ME            : E1 Li mode (default)\n");
  DemoLogf(" /MT            : T1 Li mode\n");
}



static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);

}



int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_INT32 i;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  for (i = 0; i < 4; i++) {
    g_volume.sqCodec[i] = 50;  // Volume can be set between 0 and 100. 0 turns the sound completely off   
    g_hPipes.hFromT1E1[i]           = OTX_INVALID_HANDLE_VALUE;
    g_hPipes.bT1E1Connected[i]      = OTX_FALSE;
    g_hPipes.hFromAudioVolumeTs[i]  = OTX_INVALID_HANDLE_VALUE;
    g_hPipes.bAVConnected[i]        = OTX_FALSE;
  }

  DemoLogf("AudioVolume, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Audio Volume Demo Program\nfor Odin's Thor-1-PCI-Basic/Vidar-5x1-ASM-Basic board combo\n");
  DemoLogf("and Sleipinr-PCI-Plus-2.0 board\n");

  // Set the default values and parse command line arguments
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Connect to the Driver
  DemoLogf("Connecting lib...");
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board (important to set the clock source before configuring the E1/T1 devices)
  result = OtxBrdSetClocks(g_hPhysDevs.hBoardController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // The following code is needed to make sure that the DSP is synchronized to the frame sync signal for the highways
  {  
    OTX_INT32 nDsp;
    for (nDsp = 0; nDsp < DEMO_DSP_MAX; nDsp++) {
      DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
    }
  }


  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Create pipes and connect the logical devices...\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Load the default values to the AudioVolume application
  DemoLoadVolumeValues();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling all opened logical devices...\n");
  DemoEnableLogicalDevices();

  // Start thread for status events from the driver 
  OsaThreadStart(DemoIdleFunction);
  OsaThreadSleep(300); // To let DemoIdleFunction start up

  // Which version of driver? Note version 4 was the first one to 
  // provide this feature. Debug Viewer will display date driver
  // was built.
  result = OtxDrvDbgDump(g_hPhysDevs.hBoardController0, 0x4002);
  if(OTX_FAILURE(result)) {
    DemoLogf("\nDriver release 1, 2 or 3\n");
  }
  else {
    // Result code has the release number
    DemoLogf("\nDriver release %d\n", result & 0xFF);
  }

  // Main thread. Wait for commands from the user ....
  DemoLogf("\nMain loop running. Press 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  DemoExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/ME", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
